Businesses will have to provide the following information to internet users when seeking their consent.
Who is collecting the data, and how to contact them or their European representative.
What the personal information are being used for, and the legal basis of the data processing.
The “legitimate interest” of the user of the data This refers to a legal basis that may be used by direct marketing companies).
With whom the data will be shared.
Whether the controller intends to transfer data to a third country, and if so has the European Commission deemed this country’s protections adequate or what alternative safeguards or rules are in place.
The duration of storage, or the criteria used to determine duration.
That the user has the right to request rectification to mistakes in this personal information.
That the user has the right to withdraw consent.
How the user can lodge a complaint with the supervisory authority.
What the consequences of not giving consent might be.
In cases of automated decision-making, including profiling, what the logic of this process is, and what the significance of the outcomes may be.
